from django.db import connections, transaction
from django.db.models import Q
from django.db.models.sql.constants import (
    GET_ITERATOR_CHUNK_SIZE,
    NO_RESULTS,
)
from django.conf import settings
from django.db.models.sql.where import WhereNode


from .utils import (
    get_current_tenant,
    get_tenant_filters,
    is_distributed_model,
)


def add_tenant_filters_on_query(obj):
    current_tenant = get_current_tenant()

    if current_tenant:
        try:
            filters = get_tenant_filters(obj.model)
            obj.add_q(Q(**filters))
        except ValueError:
            pass


def wrap_get_compiler(base_get_compiler):
    # Adds tenant filters to the query object
    def get_compiler(obj, *args, **kwargs):
        add_tenant_filters_on_query(obj)
        return base_get_compiler(obj, *args, **kwargs)

    # pylint: disable=protected-access
    get_compiler._sign = "get_compiler django-multitenant"
    return get_compiler


# pylint: disable=unused-argument
def wrap_update_batch(base_update_batch):
    # Written to decorate the update_batch method of the UpdateQuery class to add tenant_id filters.
    # Since add tenant_filters method must be executed before the execute_sql method, we have to
    # copy and rewrite the update_batch method.
    # CAUTION: Since source is copied, UpdateQuery.update_batch method should be tracked
    def update_batch(obj, pk_list, values, using):
        obj.add_update_values(values)
        for offset in range(0, len(pk_list), GET_ITERATOR_CHUNK_SIZE):
            obj.where = WhereNode()
            obj.add_q(Q(pk__in=pk_list[offset : offset + GET_ITERATOR_CHUNK_SIZE]))
            add_tenant_filters_on_query(obj)
            obj.get_compiler(using).execute_sql(NO_RESULTS)

    # pylint: disable=protected-access
    update_batch._sign = "update_batch django-multitenant"
    return update_batch


def wrap_delete(base_delete):
    def delete(obj):
        obj_are_distributed = [
            is_distributed_model(instance) for instance in obj.data.keys()
        ]

        # If all elements are from distributed tables, then we can do a simple atomic transaction.
        # If there are mixes of distributed and reference tables, it would raise the following error :

        # django.db.utils.InternalError: cannot execute DML on reference relation
        # "table" because there was a parallel DML access to distributed relation
        # "table2" in the same transaction

        if len(set(obj_are_distributed)) > 1 and getattr(
            settings, "CITUS_EXTENSION_INSTALLED", False
        ):
            # all elements are not the same, some False, some True
            with transaction.atomic(using=obj.using, savepoint=False):
                connections[obj.using].cursor().execute(
                    "SET LOCAL citus.multi_shard_modify_mode TO 'sequential';"
                )
                result = base_delete(obj)
                connections[obj.using].cursor().execute(
                    "SET LOCAL citus.multi_shard_modify_mode TO 'parallel';"
                )
                return result

        return base_delete(obj)

    return delete



# pylint: disable=unused-import
import django_multitenant.django_mt_environment
from django_multitenant.utils import set_current_tenant
from django_multitenant.models import TenantModel
from rest_framework import viewsets
from abc import abstractmethod


@abstractmethod
def get_tenant(request) -> TenantModel:
    pass


class TenantModelViewSet(viewsets.ModelViewSet):
    model_class = TenantModel

    def get_queryset(self):
        if self.request.user.is_anonymous:
            return self.model_class.objects.none()
        tenant = get_tenant(self.request)
        set_current_tenant(tenant)
        list3 = self.model_class.objects.all()
        return list3



from django_multitenant.utils import set_current_tenant

from django_multitenant.views import get_tenant


class MultitenantMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        if request.user and not request.user.is_anonymous:
            tenant = get_tenant(request)
            set_current_tenant(tenant)
        return self.get_response(request)



import logging
import django
from django.db import models
from django.db.models.sql.where import WhereNode
from django.conf import settings

from .exceptions import EmptyTenant
from .utils import get_current_tenant, get_tenant_column, get_tenant_filters

logger = logging.getLogger(__name__)


class TenantForeignKey(models.ForeignKey):
    """
    Should be used in place of models.ForeignKey for all foreign key relationships to
    subclasses of TenantModel.

    Adds additional clause to JOINs over this relation to include tenant_id in the JOIN
    on the TenantModel.

    Adds clause to forward accesses through this field to include tenant_id in the
    TenantModel lookup.
    """

    # Override
    def get_extra_descriptor_filter(self, instance):
        """
        Return an extra filter condition for related object fetching when
        user does 'instance.fieldname', that is the extra filter is used in
        the descriptor of the field.

        The filter should be either a dict usable in .filter(**kwargs) call or
        a Q-object. The condition will be ANDed together with the relation's
        joining columns.

        A parallel method is get_extra_restriction() which is used in
        JOIN and subquery conditions.
        """

        current_tenant = get_current_tenant()
        if current_tenant:
            return get_tenant_filters(self.related_model)

        empty_tenant_message = (
            f"TenantForeignKey field {self.model.__name__}.{self.name} "
            "accessed without a current tenant set. "
            "This may cause issues in a partitioned environment. "
            "Recommend calling set_current_tenant() before accessing "
            "this field."
        )

        if getattr(settings, "TENANT_STRICT_MODE", False):
            raise EmptyTenant(empty_tenant_message)

        logger.warning(empty_tenant_message)
        return super().get_extra_descriptor_filter(instance)

    # Override
    # Django 4.0 removed the where_class argument from this method, so
    # depending on the version we define the function with a different
    # signature.
    # pylint: disable=unused-argument,arguments-differ

    if django.VERSION >= (4, 0):

        def get_extra_restriction(self, alias, related_alias):
            return self.get_extra_restriction_citus(alias, related_alias)

    else:

        def get_extra_restriction(self, where_class, alias, related_alias):
            return self.get_extra_restriction_citus(alias, related_alias)

    def get_extra_restriction_citus(self, alias, related_alias):
        """
        Return a pair condition used for joining and subquery pushdown. The
        condition is something that responds to as_sql(compiler, connection)
        method.

        Note that currently referring both the 'alias' and 'related_alias'
        will not work in some conditions, like subquery pushdown.

        A parallel method is get_extra_descriptor_filter() which is used in
        instance.fieldname related object fetching.
        """

        if not (related_alias and alias):
            return None

        # Fetch tenant column names for both sides of the relation
        lhs_model = self.model
        rhs_model = self.related_model
        lhs_tenant_id = get_tenant_column(lhs_model)
        rhs_tenant_id = get_tenant_column(rhs_model)

        # Fetch tenant fields for both sides of the relation
        lhs_tenant_field = lhs_model._meta.get_field(lhs_tenant_id)
        rhs_tenant_field = rhs_model._meta.get_field(rhs_tenant_id)

        # Get references to both tenant columns
        lookup_lhs = lhs_tenant_field.get_col(related_alias)
        lookup_rhs = rhs_tenant_field.get_col(alias)

        # Create "AND lhs.tenant_id = rhs.tenant_id" as a new condition
        lookup = lhs_tenant_field.get_lookup("exact")(lookup_lhs, lookup_rhs)
        condition = WhereNode()
        condition.add(lookup, "AND")
        return condition


class TenantOneToOneField(models.OneToOneField, TenantForeignKey):
    # Override
    def __init__(self, *args, **kwargs):
        kwargs["unique"] = False
        super().__init__(*args, **kwargs)



import logging

from django.db.models.sql import DeleteQuery, UpdateQuery
from django.db.models.deletion import Collector
from django.db.utils import NotSupportedError
from django.conf import settings

import django
from django.db.models.fields.related_descriptors import (
    create_forward_many_to_many_manager,
)


from .deletion import related_objects
from .exceptions import EmptyTenant
from .query import wrap_get_compiler, wrap_update_batch, wrap_delete
from .utils import (
    set_current_tenant,
    get_current_tenant,
    get_current_tenant_value,
    get_tenant_field,
    get_tenant_filters,
    get_object_tenant,
    set_object_tenant,
    get_tenant_column,
)


logger = logging.getLogger(__name__)


def wrap_many_related_manager_add(many_related_manager_add):

    """
    Wraps the add method of many to many field to set tenant_id in through_defaults
    parameter of the add method.
    """

    def add(self, *objs, through_defaults=None):

        if hasattr(self.through, "tenant_field") and get_current_tenant():
            through_defaults = through_defaults or {}
            through_defaults[
                get_tenant_column(self.through)
            ] = get_current_tenant_value()
        return many_related_manager_add(self, *objs, through_defaults=through_defaults)

    return add


def wrap_forward_many_to_many_manager(create_forward_many_to_many_manager_method):

    """
    Wraps the create_forward_many_to_many_manager method of the related_descriptors module
    and changes the add method of the ManyRelatedManagerClass to set tenant_id in through_defaults
    """

    def create_forward_many_to_many_manager_wrapper(superclass, rel, reverse):
        ManyRelatedManagerClass = create_forward_many_to_many_manager_method(
            superclass, rel, reverse
        )
        ManyRelatedManagerClass.add = wrap_many_related_manager_add(
            ManyRelatedManagerClass.add
        )
        return ManyRelatedManagerClass

    # pylint: disable=protected-access
    create_forward_many_to_many_manager_wrapper._sign = "add django-multitenant"
    return create_forward_many_to_many_manager_wrapper


class TenantManagerMixin:
    # Below is the manager related to the above class.
    # Overrides the get_queryset method of to inject tenant_id filters in the get_queryset.
    # With this method, models extended from TenantManagerMixin will have tenant_id filters by default
    # if the tenant_id is set in the current thread.
    def get_queryset(self):
        # Injecting tenant_id filters in the get_queryset.
        # Injects tenant_id filter on the current model for all the non-join/join queries.
        queryset = self._queryset_class(self.model)
        current_tenant = get_current_tenant()
        if current_tenant:
            kwargs = get_tenant_filters(self.model)
            return queryset.filter(**kwargs)
        return queryset

    def bulk_create(self, objs, **kwargs):
        # Helper method to set tenant_id in the current thread for the results returned from query_set.
        # For example, if we have a query_set of all the users in the current tenant, we can set the tenant_id by calling
        # User.object.bulk_create(users)
        if get_current_tenant():
            tenant_value = get_current_tenant_value()
            for obj in objs:
                set_object_tenant(obj, tenant_value)

        return super().bulk_create(objs, **kwargs)


class TenantModelMixin:
    # Abstract model which all the models related to tenant inherit.

    def __init__(self, *args, **kwargs):
        # Adds tenant_id filters in the delete and update queries.

        # Below block decorates Delete operations related activities and adds tenant_id filters.

        if not hasattr(DeleteQuery.get_compiler, "_sign"):
            # Decorates the the compiler of DeleteQuery to add tenant_id filters.
            DeleteQuery.get_compiler = wrap_get_compiler(DeleteQuery.get_compiler)
            # Decorates related_objects to add tenant_id filters to add tenant_id filters.
            # related_objects is being used to define additional records for deletion defined with relations
            Collector.related_objects = related_objects
            # Decorates the delete method of Collector to execute citus shard_modify_mode commands
            # if distributed tables are being related to the model.
            Collector.delete = wrap_delete(Collector.delete)

        if not hasattr(create_forward_many_to_many_manager, "_sign"):
            django.db.models.fields.related_descriptors.create_forward_many_to_many_manager = wrap_forward_many_to_many_manager(
                create_forward_many_to_many_manager
            )

        # Decorates the update_batch method of UpdateQuery to add tenant_id filters.
        if not hasattr(UpdateQuery.get_compiler, "_sign"):
            UpdateQuery.update_batch = wrap_update_batch(UpdateQuery.update_batch)

        super().__init__(*args, **kwargs)

    def __setattr__(self, attrname, val):
        # Provides failing of the save operation if the tenant_id is changed.
        # try_update_tenant is being checked inside save method and if it is true, it will raise an exception.
        def is_val_equal_to_tenant(val):
            return (
                val
                and self.tenant_value
                and val != self.tenant_value
                and val != self.tenant_object
            )

        if (
            attrname in (self.tenant_field, get_tenant_field(self).name)
            and not self._state.adding
            and is_val_equal_to_tenant(val)
        ):
            self._try_update_tenant = True

        return super().__setattr__(attrname, val)

    # pylint: disable=too-many-arguments
    def _do_update(self, base_qs, using, pk_val, values, update_fields, forced_update):
        # adding tenant filters for save
        # Citus requires tenant_id filters for update, hence doing this below change.

        current_tenant = get_current_tenant()

        if current_tenant:
            kwargs = get_tenant_filters(self.__class__)
            base_qs = base_qs.filter(**kwargs)
        else:
            empty_tenant_message = (
                f"Attempting to update {self._meta.model.__name__} instance {self} "
                "without a current tenant set. "
                "This may cause issues in a partitioned environment. "
                "Recommend calling set_current_tenant() before performing this "
                "operation.",
            )
            if getattr(settings, "TENANT_STRICT_MODE", False):
                raise EmptyTenant(empty_tenant_message)
            logger.warning(empty_tenant_message)

        return super()._do_update(
            base_qs, using, pk_val, values, update_fields, forced_update
        )

    def save(self, *args, **kwargs):
        # Performs tenant related operations before and after save.
        # Synchronizes object tenant with the tenant set in the application in case of a mismatch
        # In normal cases _try_update_tenant should prevent tenant_id from being updated.
        # However, if the tenant_id is updated in the database directly, this will catch it.
        if hasattr(self, "_try_update_tenant"):
            raise NotSupportedError("Tenant column of a row cannot be updated.")

        current_tenant = get_current_tenant()
        tenant_value = get_current_tenant_value()

        set_object_tenant(self, tenant_value)

        if self.tenant_value and tenant_value != self.tenant_value:
            self_tenant = get_object_tenant(self)
            set_current_tenant(self_tenant)

        try:
            obj = super().save(*args, **kwargs)
        finally:
            set_current_tenant(current_tenant)

        return obj

    @property
    def tenant_field(self):
        if hasattr(self, "TenantMeta") and "tenant_field_name" in dir(self.TenantMeta):
            return self.TenantMeta.tenant_field_name
        if hasattr(self, "TenantMeta") and "tenant_id" in dir(self.TenantMeta):
            return self.TenantMeta.tenant_id
        if hasattr(self, "tenant"):
            raise AttributeError(
                f"Tenant field exists which may cause collision with tenant_id field. Please rename the tenant field in {self.__class__.__name__} "
            )
        if hasattr(self, "tenant_id"):
            return self.tenant_id
        if self.__module__ == "__fake__":
            raise AttributeError(
                f"apps.get_model method should not be used to get the model {self.__class__.__name__}."
                "Either import the model directly or use the module apps under the module django.apps."
            )

        raise AttributeError(
            f"tenant_id field not found. Please add tenant_id field to the model {self.__class__.__name__}"
        )

    @property
    def tenant_value(self):
        return getattr(self, self.tenant_field, None)

    @property
    def tenant_object(self):
        return get_object_tenant(self)



import operator
from functools import reduce

import django
from django.db.models import Q

from .utils import get_current_tenant, get_tenant_filters


def related_objects(obj, *args):
    """
    Override of Collector.related_objects. Returns the filter for the related objects
    Different from the original method, this method adds the tenant filters to the query.
    CAUTION: When Collector.related_objects is changed, this method should be updated accordingly.
    """
    if django.VERSION < (3, 0) or len(args) == 2:
        related = args[0]
        related_model = related.related_model
        related_fields = [related.field]
        objs = args[1]
    else:
        # Starting django 3.1 the signature of related_objects changed to
        # def related_objects(self, related_model, related_fields, objs)
        related_model = args[0]
        related_fields = args[1]
        objs = args[2]

    filters = {}
    predicate = reduce(
        operator.or_,
        (Q(**{f"{related_field.name}__in": objs}) for related_field in related_fields),
    )

    if get_current_tenant():
        try:
            filters = get_tenant_filters(related_model)
        except ValueError:
            pass
    # pylint: disable=protected-access
    return related_model._base_manager.using(obj.using).filter(predicate, **filters)



import logging
import django
from django.db.backends.postgresql.base import (
    DatabaseFeatures as PostgresqlDatabaseFeatures,
    DatabaseWrapper as PostgresqlDatabaseWrapper,
    DatabaseSchemaEditor as PostgresqlDatabaseSchemaEditor,
)
from django_multitenant.fields import TenantForeignKey
from django_multitenant.utils import get_model_by_db_table, get_tenant_column

logger = logging.getLogger(__name__)


class DatabaseSchemaEditor(PostgresqlDatabaseSchemaEditor):
    sql_create_column_inline_fk = None

    # Override
    def __enter__(self):
        ret = super().__enter__()
        return ret

    # pylint: disable=too-many-arguments
    def _alter_field(
        self,
        model,
        old_field,
        new_field,
        old_type,
        new_type,
        old_db_params,
        new_db_params,
        strict=False,
    ):
        """
        If there is a change in the field, this method assures that if the field is type of TenantForeignKey
        and db_constraint does not exist, adds the foreign key constraint.
        """

        super()._alter_field(
            model,
            old_field,
            new_field,
            old_type,
            new_type,
            old_db_params,
            new_db_params,
            strict,
        )

        # If the pkey was dropped in the previous distribute migration,
        # foreign key constraints didn't previously exists so django does not
        # recreated them.
        # Here we test if we are in this case
        if isinstance(new_field, TenantForeignKey) and new_field.db_constraint:
            from_model = get_model_by_db_table(model._meta.db_table)
            fk_names = self._constraint_names(
                model, [new_field.column], foreign_key=True
            ) + self._constraint_names(
                model,
                [new_field.column, get_tenant_column(from_model)],
                foreign_key=True,
            )
            if not fk_names:
                self.execute(
                    self._create_fk_sql(
                        model, new_field, "_fk_%(to_table)s_%(to_column)s"
                    )
                )

    # Override
    def _create_fk_sql(self, model, field, suffix):
        """
        This method overrides the additions foreign key constraint sql and adds the tenant column to the constraint
        """
        if isinstance(field, TenantForeignKey):
            try:
                # test if both models exists
                # This case happens when we are running from scratch migrations and one model was removed from code
                # In the previous migrations we would still be creating the foreign key
                from_model = get_model_by_db_table(model._meta.db_table)
                to_model = get_model_by_db_table(
                    field.target_field.model._meta.db_table
                )
            except ValueError:
                return None

            from_columns = field.column, get_tenant_column(from_model)
            to_columns = field.target_field.column, get_tenant_column(to_model)
            suffix = suffix % {
                "to_table": field.target_field.model._meta.db_table,
                "to_column": "_".join(to_columns),
            }

            return self.sql_create_fk % {
                "table": self.quote_name(model._meta.db_table),
                "name": self.quote_name(
                    self._create_index_name(
                        model._meta.db_table, from_columns, suffix=suffix
                    )
                ),
                "column": ", ".join(
                    [self.quote_name(from_col) for from_col in from_columns]
                ),
                "to_table": self.quote_name(field.target_field.model._meta.db_table),
                "to_column": ", ".join(
                    [self.quote_name(to_col) for to_col in to_columns]
                ),
                "deferrable": self.connection.ops.deferrable_sql(),
            }
        return super()._create_fk_sql(model, field, suffix)

    # Override
    def execute(self, sql, params=()):
        # Hack: Citus will throw the following error if these statements are
        # not executed separately: "ERROR: cannot execute multiple utility events"
        if sql and not params:
            for statement in str(sql).split(";"):
                if statement and not statement.isspace():
                    super().execute(statement)
        elif sql:
            super().execute(sql, params)


# noqa
class TenantDatabaseFeatures(PostgresqlDatabaseFeatures):
    # The default Django behaviour is to collapse the fields to just the 'id'
    # field. This doesn't work because we're using a composite primary key. In
    # Django version 3.0 a function was added that we can override to specify
    # for specific models that this behaviour should be disabled.
    def allows_group_by_selected_pks_on_model(self, model):
        # pylint: disable=import-outside-toplevel
        from django_multitenant.models import TenantModel

        if issubclass(model, TenantModel):
            return False
        return super().allows_group_by_selected_pks_on_model(model)

    # For django versions before version 3.0 we set a flag that disables this
    # behaviour for all models.
    if django.VERSION < (3, 0):
        allows_group_by_selected_pks = False


class DatabaseWrapper(PostgresqlDatabaseWrapper):
    # Override
    SchemaEditorClass = DatabaseSchemaEditor
    features_class = TenantDatabaseFeatures



import importlib


def import_drf_or_die():
    try:
        importlib.import_module("rest_framework")
    except ImportError as e:
        raise ImportError(
            "Django Multitenant requires Django Rest Framework to be installed."
        ) from e


import_drf_or_die()



from django.conf import settings

APP_NAMESPACE = "DJANGO_MULTITENANT"

TENANT_COOKIE_NAME = getattr(settings, "TENANT_COOKIE_NAME", "tenant_id")
TENANT_MODEL_NAME = getattr(settings, "TENANT_MODEL_NAME", None)
CITUS_EXTENSION_INSTALLED = getattr(settings, "CITUS_EXTENSION_INSTALLED", False)
TENANT_STRICT_MODE = getattr(settings, "TENANT_STRICT_MODE", False)



from setuptools import setup, find_packages
from os import path
from io import open

# read the contents of your README file
this_directory = path.dirname(path.abspath(__file__))

with open(path.join(this_directory, "README.md")) as f:
    long_description = f.read()


# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.

setup(
    name="django-multitenant",
    version="3.2.1",  # Required
    description="Django Library to Implement Multi-tenant databases",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/citusdata/django-multitenant",
    author="Louise Grandjonc",
    author_email="louise.grandjonc@microsoft.com",
    # Classifiers help users find your project by categorizing it.
    #
    # For a list of valid classifiers, see https://pypi.org/classifiers/
    classifiers=[
        "Development Status :: 5 - Production/Stable ",
        "Topic :: Database",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    keywords=("citus django multi tenant" "django postgres multi-tenant"),
    packages=find_packages(
        exclude=["*.tests", "*.tests.*", "tests.*", "tests", "docs", "docs.*"]
    ),
)



# -*- coding: utf-8 -*-
import django

version = (2, 3, 2)

__version__ = ".".join(map(str, version))

# default_app_config is auto detected for versions 3.2 and higher:
# https://docs.djangoproject.com/en/3.2/ref/applications/#for-application-authors
if django.VERSION < (3, 2):
    default_app_config = "django_multitenant.apps.MultitenantConfig"

__all__ = ["default_app_config", "version"]
